% MPC for ZCU FEL - OFFLINE Version (without communication to REX)
%
%   1 - Number of networks
%   1 - Renewable source (RES) - solar + wind
%   1 - Battery (BESS)
%
%   x(k+1) = A*x(k) + Bu*u(k) + Bv*v(k)
%   y(k)   = C*x(k)
%
%   x0 = 0
%
%   State vector:
%       x(k) = [P(k), T(k), S(k)]'   3 x 3
%   Input vector:
%       u(k) = [difP(k), difT(k)]'   2 x 2
%   Disturbance vector:
%       v(k) = [PL(k), PR(k)]'       2 x 1
%   Output vector:
%       y(k) = [Pimbal(k)]           1 x 1
%
%   Charging:
%       alpha: addition coefficient
%       Sref:  State of Charge reference value
%
%   Balancing:
%       Pimref: Balancing setpoint
%
%   Cost of network power: 
%       cost = ca*Pi + cb(i)
%   Cost vector 1 x 2:
%       [ca, cb]
% 
%   Cost function weights
%       w1 ... weight of Pimbal square
%       w2 ... weight of sum of cost squares
%       w3 ... weight of square of (S(k)-Sref)
%       w4 ... weight of sum of input squares
%
%   MPC Horizons
%       np ... prediction horizon
%       nc ... control horizon
%
%   Simulation parameters
%       nsteps ... number of simulation steps
%       PL ... Load power (disturbance)                     nsteps x 1
%       PR ... Renewable power (disturbance)                nsteps x 1
%--------------------------------------------------------------------------

% Initial parameters
Ts = 1/12;  % [h], 5 min
%Ts = 2/12; % [h], 10 min
Emax = 208; % [kWh], maximum BESS capacity 62.4
alpha = Ts/Emax;
% % % Sref = 0.7; %0.5
% % % Smin = 0.5; %0.2
% % % Smax = 0.98; %0.9
Sref = 0.9; %0.5
Smin = 0.5; %0.2
Smax = 0.98; %0.9

% % x0 = [0; 0; 0.5*(Smax+Smin)]; puvodni nabiti baterie na zacatku
% simulace
x0 = [0; 0; Sref];
x = x0;

Pmax = 200; % kW
Pinv = 85;  % kW 
Pimref = 0.0;

% Constraints
xmin = [0.0; -1e3; Smin];
xmax = [Pmax;+1e3; Smax];

% cost: ca*P*P + cb*P + cc in CZK
c = [0; 4; 0];     % ca = 0, cb = 4 Kc/kWh, cc = 0

% % %  % Weights puvodni 
% % % w1 = 0;    % Pimbal
% % % w2 = 0.1;  % Cost
% % % w3 = 100;    % ESoC
% % % w4 = 0.5;    % U

 % Weights puvodni 
w1 = 0;    % Pimbal
w2 = 0.1;  % Cost
w3 = 100;    % ESoC
w4 = 0.5;    % U

% % o	1. pvodn  0:0.1:100:0.5
% % o	2. stejn  0:1:1:1
% % o	3. reln  0:0.1:10:0.5


% % % Weights nove
% % w1 = 0;    % Pimbal
% % w2 = 1;    % Cost
% % w3 = 1;    % ESoC
% % w4 = 1;    % U

% MPC Horizons
np = 144; %144
nc = np;

% Prepare data
[nsteps,PL,PR] = prepare_data(np);

% Set reasonable x0
x0(1) = (PL(1)-PR(1));

V = zeros(2*np,1);
PLmR = zeros(np,1);
for kk = 1:np
    V(2*kk-1:2*kk,1) = [PL(kk);PR(kk)];
    PLmR(kk) = PL(kk) - PR(kk);
end
Xmin = zeros(3*np,1);
Xmax = Xmin;

%--------------------------------------------------------------------------
% Simulation
XY = zeros(nsteps,5);
SumPimbalSqr = zeros(nsteps-np+1,1);
SumCostSqr = zeros(nsteps-np+1,1);
SumESSqr = zeros(nsteps-np+1,1);
SumUSqr = zeros(nsteps-np+1,1);
PS = zeros(nsteps,1);
Jval = zeros(nsteps,1);   % values of cost functions for all simulation steps
Jexits = zeros(nsteps,1); % quad prog exitflags
Utmp = zeros(2*np,1);

disp(sprintf('\n'));
for k = 1:nsteps
    opts1 = optimset('Diagnostics','off');
    warning off verbose

    [err,U,x,y,exitflag,Jval(k),Jparts] = ... 
        MPC_quadprog(np,nc,x0,alpha,Sref,Pimref, ...
        [w1;w2;w3;w4],c,0,Pmax,Smin,Smax,Pinv,PLmR,V,Ts);
    
    %optimset('display',opts1);
    disp(sprintf('\b%d,',k));
    if (mod(k,10) == 0)
        disp(sprintf(' \b'));
    end
    Jexits(k) = exitflag;
    % exitflag values
    %    1  QUADPROG converged with a solution X.
    %    3  Change in objective function value smaller than the specified tolerance.
    %    4  Local minimizer found.
    %    0  Maximum number of iterations exceeded.
    %   -2  No feasible point found.
    %   -3  Problem is unbounded.
    %   -4  Current search direction is not a direction of descent
    %   -7  Magnitude of search direction became too small; no further progress can be made.
    
    % Store x, y, copt
    copt = (x(1)*c(1) + c(2))*x(1) + c(3);
    XY(k,:) = [x' y copt];
    PS(k) = PLmR(1)-x(2);   % PL - PR - Tk
    % Cost function calculation (running)
    SumPimbalSqr(k) = w1*Jparts(1);
    SumCostSqr(k) = w2*Jparts(2);
    SumESSqr(k) = w3*Jparts(3);
    SumUSqr(k) = w4*Jparts(4);
     
    % Prepare next iteration
    x0 = x;
    V = [V(3:2*np); PL(k+np); PR(k+np)];
    PLmR = [PLmR(2:np); PL(k+np)- PR(k+np)];
end % for k

J = [w1*SumPimbalSqr(1:nsteps) w2*SumCostSqr(1:nsteps) w3*SumESSqr(1:nsteps) w4*SumUSqr(1:nsteps)...
    (w1*SumPimbalSqr(1:nsteps)+w2*SumCostSqr(1:nsteps)+w3*SumESSqr(1:nsteps)+w4*SumUSqr(1:nsteps))];

Plr=[PL(1:nsteps) PR(1:nsteps)];

save('./DATA/Plr.txt','Plr','-ascii');
save('./DATA/XY.txt','XY','-ascii');
save('./DATA/J.txt','J','-ascii');

subplot(411)
plot(1:nsteps, XY(:,4),'-',1:nsteps,XY(:,1),'-', 1:nsteps,XY(:,2),'-', 1:nsteps,PS,'-', 1:nsteps,XY(:,3),'-');
legend('Pimbal','P','T','PS','SoC');

subplot(412)
plot(1:nsteps,PL(1:nsteps),'-',1:nsteps,PR(1:nsteps),'-');
legend('PL','PR');

subplot(413)
plot(1:nsteps,J(:,1),'-',1:nsteps, J(:,2),'--',1:nsteps, J(:,3),':',1:nsteps, J(:,4),'-.');
legend('w1*Pimbal','w2*Cost','w3*ESoC','w4*U');

subplot(414)
plot(1:nsteps,XY(:,3),'-');
legend('SoC');

Jval
Jexits
J
