% MPC quadprog for ZCU FEL
%
% Assemble MPC matrices with respect to the selected generators
%   and solve QP problem
%
%   1           number of networks (simulated as "big" generator
%   1           Renewable source (RES)
%   1           Battery (BESS)
%
%   np          prediction horizon
%   nc          control horizon
%
%   x0          initial state is a vector 3 x 1 
%
%   alpha       Charging addition coefficient
%   Sref        State of Charge reference value
%   Pimref      Setpoint of balancing generator
%
%   w           vector 4 x 1 of cost function weights
%       w(1)    weight of Pimbal square
%       w(2)    weight of sum of cost squares
%       w(3)    weight of square of (S(k)-Sref)
%       w(4)    weight of sum of input squares
%
%   Cost of network power P: 
%       cost = c(1)*P*P + c(2)*P + c(3)
% 
%   Pmin,Pmax   minimum and maximum network powers
%   Smin,Smax   minimum and maximum of the battery State of charge
%   V           matrix 2*np x 1 of predicted load and renewables powers
%   PLmR        vector np x 1 of vectors PL - PR with np x 1 elements
%   Pinv        maximum power of solar invertor
%	PLres		load reserve
%
function [err,U,x,y,exitflag,J,Jparts] = ...
         MPC_quadprog(np,nc,x0,alpha,Sref,Pimref,...
            w,c,Pmin,Pmax,Smin,Smax,Pinv,PLmR,V,Ts)
%
%   State vector:
%       x(k) = [P(k), T(k), S(k)]'   3 x 3
%   Input vector:
%       u(k) = [difP(k), difT(k)]'   2 x 2
%   Disturbance vector:
%       v(k) = [PL(k), PR(k)]'       2 x 1
%   Output vector:
%       y(k) = [Pimbal(k)]           1 x 1
if (np < nc)
    error('ERROR: control horizon nc must be less or equal to prediction horizon np');
end
err = 0;
% nsel is 1 and ns = 3 always

% Initiate model matrices
A = eye(3); A(3,2) = alpha;
Bu = eye(3,2);
Bv = [zeros(2,2); -alpha alpha];
C = [-1 1 0];
CT = [0 1 0];

% Construct Cost function matrices
Q1 = [1 -1 0; -1 1 0; 0 0 0];
Q2 = [c(1) 0 0; 0 0 0; 0 0 0];
Q3 = [0 0 0; 0 0 0; 0 0 1];
Q = w(1)*Q1 + w(2)*Q2 + w(3)*Q3;
W = [2*Pimref*w(1)+w(2)*c(2) -2*Pimref*w(1) -2*Sref*w(3)];
R = [w(4) 0; 0 0]; 

% Construct large matrices Qnp, Wnp, Rnp
Qnp = zeros(np*3);
Wnp = zeros(1,np*3);
Rnp = zeros(np*2);

for i = 1:np
    Qnp((i-1)*3+1:i*3, (i-1)*3+1:i*3) = Q;
    Wnp(1, (i-1)*3+1:i*3) = W;
    Rnp((i-1)*2+1:i*2, (i-1)*2+1:i*2) = R;
end

% Construct large matrices SUnp, SVnp, Tnp, SCUnp, SCVnp, TCnp, STUnp, TTnp
SUnp = zeros(np*3,np*2);
SVnp = zeros(np*3,np*2);
Tnp = zeros(np*3,3);
SUnp(1:3,1:2) = Bu;
SVnp(1:3,1:2) = Bv;
Tnp(1:3,:) = A;

SCUnp = zeros(nc*1,nc*2);
SCVnp = zeros(nc*1,nc*2);
TCnp = zeros(nc*1,3);
SCUnp(1,1:2) = C*Bu;
SCVnp(1,1:2) = C*Bv;
TCnp(1,:) = C;

STUnp = zeros(nc*1,nc*2);
TTnp = zeros(nc*1,3);
STUnp(1,1:2) = CT*Bu;
TTnp(1,:) = CT;

for i = 2:np
    Tnp((i-1)*3+1:i*3,:) = Tnp((i-2)*3+1:(i-1)*3,:);
    Tnp(i*3,2) = i*alpha;
    if (i <= nc)
        TCnp(i,:) = TCnp(i-1,:);
        TTnp(i,:) = TTnp(i-1,:);
    end
    for j = i:-1:2
        SUnp((i-1)*3+1:i*3,(j-1)*2+1:j*2) = SUnp((i-2)*3+1:(i-1)*3,(j-2)*2+1:(j-1)*2);
        SVnp((i-1)*3+1:i*3,(j-1)*2+1:j*2) = SVnp((i-2)*3+1:(i-1)*3,(j-2)*2+1:(j-1)*2);

        if (i <= nc)
            SCUnp(i,(j-1)*2+1:j*2) = SCUnp(i-1,(j-2)*2+1:(j-1)*2);
            SCVnp(i,(j-1)*2+1:j*2) = SCVnp(i-1,(j-2)*2+1:(j-1)*2);

            STUnp(i,(j-1)*2+1:j*2) = STUnp(i-1,(j-2)*2+1:(j-1)*2);
        end
    end
    SUnp((i-1)*3+1:i*3,1:2) = SUnp((i-2)*3+1:(i-1)*3,1:2);
    SUnp(i*3,2) = (i-1)*alpha; 
    SVnp((i-1)*3+1:i*3,1:2) = SVnp((i-2)*3+1:(i-1)*3,1:2);
    
    if (i <= nc)
        SCUnp(i,1:2) = SCUnp(i-1,1:2);
        %SCVnp(i,1:2) = [0 0];
        STUnp(i,1:2) = STUnp(i-1,1:2);
    end
end

% Prepare other matrices
L = eye(np*2, nc*2);
% in previous versions was error. Our criterion is min x'*H*x + f'*x
% but quadprog function requires min 0.5*x'*H*x + f'*x   subject to:  A*x <= b
H = 2*L'*(SUnp'*Qnp*SUnp + Rnp)*L;  % now it is corrected by multiplication 2*
H = 0.5*(H+H');     % be sure that H is symmetric
Aqp = [SUnp*L; -SUnp*L; STUnp; -STUnp];
Aeq = [SCUnp];
x = zeros(3,1);    % full size as x0 
xmin = [Pmin -1e3 Smin];
xmax = [Pmax  1e3 Smax];

Xmin = zeros(3*np,1);
Xmax = Xmin;
for i = 1:np
    Xmin((i-1)*3+1:i*3) = xmin;
    Xmax((i-1)*3+1:i*3) = xmax;
end

fT = (2*V'*SVnp'*Qnp*SUnp + 2*x0'*Tnp'*Qnp*SUnp + Wnp*SUnp)*L;
bqp = [Xmax - SVnp*V - Tnp*x0; -Xmin + SVnp*V + Tnp*x0;...
       -TTnp*x0 + PLmR(1:nc) + Pinv; TTnp*x0 - PLmR(1:nc) + Pinv];
beq = [-SCVnp*V(1:2*nc,:) - TCnp*x0];

% Call quadprog
%options = optimset('Display','off','Diagnostics','off','MaxIter',100);
options = optimset('Display','off','Diagnostics','off');
diary off
[Utmp,fval,exitflag] = quadprog(H,fT',Aqp,bqp,Aeq,beq,...
    -Inf*ones(nc*2,1),+Inf*ones(nc*2,1),zeros(nc*2,1), options);
diary on

% Prepare initialization of output variable for the case of an error
U = zeros(np*2,1);  % U is a vector for all prediction horizon
y = 0.0;
J = 0.0;
Jparts = zeros(size(w,1),1);

if (size(Utmp,1) == 0)
    err = 1;
    J = -1.0;
    return;
end

U(1:2*nc) = Utmp;

% Evaluation of the cost function
% Calculate values for whole np
PimbalSqr = zeros(1,np);
CostSqr = zeros(1,np);
ESSqr = zeros(1,np);
xtmp = x0;
Uttmp = zeros(np*2,1);  % U is a vector for all prediction horizon
Uttmp(1:nc*2) = Utmp;

for kk = 1:np
    ytmp = C*xtmp;
    PimbalSqr(kk) = (ytmp-Pimref)*(ytmp-Pimref);
    CostSqr(kk) = (c(1)*xtmp(1) + c(2))*xtmp(1) + c(3);
    ESSqr(kk) = (xtmp(3)-Sref)*(xtmp(3)-Sref);

    xtmp = A*xtmp + Bu*Uttmp((kk-1)*2+1:kk*2) + Bv*V((kk-1)*2+1:kk*2);
    ytmp = C*xtmp;
    if (kk == 1)
        x = xtmp;
        y = ytmp;
    end
end %for kk

% Cost function calculation (running)
Jparts(1) = sum(PimbalSqr);
Jparts(2) = sum(CostSqr);
Jparts(3) = sum(ESSqr);
Jparts(4) = U'*U;
J = w'*Jparts;

end %function
